import socket
import threading
import time

from objprint import add_objprint
import pyeelight


class Packet:
    pass


class OutboundRequestPacket(Packet):

    def __init__(self, man, st):
        self.man = man
        self.st = st

    def process_headers(self):
        return f"""M-SEARCH * HTTP/1.1\r\nMAN: "{self.man}"\r\nST: {self.st}\r\n""".encode()


@add_objprint
class InboundAdvertisementPacket(Packet):
    pass


class AdvertisementSocket(pyeelight.Contextable):
    MULTICAST_IP = "239.255.255.250"
    MULTICAST_PORT = 1982

    def __init__(self):
        self.running = True
        self.packets = []
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.bind((self.get_ip(), 1234))
        self.sock.settimeout(4)
        self.logger = pyeelight.Logger(self)

    def get_ip(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.settimeout(0)
        try:
            # doesn't even have to be reachable
            s.connect(('10.254.254.254', 1))
            IP = s.getsockname()[0]
        except Exception:
            IP = '127.0.0.1'
        finally:
            s.close()
        return IP

    def send_packet(self, packet: OutboundRequestPacket):
        self.sock.sendto(packet.process_headers(), (self.MULTICAST_IP, self.MULTICAST_PORT))

    def ditch(self):
        self.sock.close()

    def init_waiter_thread(self):
        self.logger.info("DISCOVERER", "Starting discovery of bulbs...")
        e = threading.Event()
        x = threading.Thread(target=self.wait_on_response)
        x.start()

        # sync memory after running thread
        x.join(3)
        e.set()

        self.running = False

        #self.ditch()
        return self.packets

    def wait_on_response(self):
        while self.running:
            try:
                data, addr = self.sock.recvfrom(1024)  # buffer size is 1024 bytes
                decoded_data = data.decode().rstrip()
                packet = InboundAdvertisementPacket()

                i = 0
                packet_dict = {}
                for element in decoded_data.split("\r\n"):
                    if i == 0:
                        packet.http = element
                    else:
                        key, value = element.split(":", 1)
                        packet_dict[key] = value.lstrip()
                    i += 1

                packet.cache_control = packet_dict["Cache-Control"]
                packet.location = packet_dict["Location"]
                packet.data = packet_dict["Date"]
                packet.ext = packet_dict["Ext"]
                packet.id = packet_dict["id"]
                packet.model = packet_dict["model"]
                packet.fw_ver = packet_dict["fw_ver"]
                packet.supported_methods = packet_dict["support"].split(" ")
                packet.power = packet_dict["power"]
                packet.brightness = packet_dict["bright"]
                packet.color_mode = packet_dict["color_mode"]
                packet.ct = packet_dict["ct"]
                packet.rgb = packet_dict["rgb"]
                packet.hue = packet_dict["hue"]
                packet.sat = packet_dict["sat"]
                packet.name = packet_dict["name"]

                self.packets.append(packet)

            except socket.timeout:
                pass

    def get_context(self):
        return f"DISCOVERER::MULTICAST"


# WIP
@add_objprint
class BulbInfo:
    def __init__(self, name, location):
        self.name = name
        self.location = location
        prefixed_address = ":".join(self.location.split(":", 2)[:2])
        self.ip = prefixed_address.split("//")[1]

    def get_ip(self):
        return self.ip

    def get_port(self):
        pass

    def get_name(self):
        return self.name

    def get_controller(self) -> pyeelight.Bulb:
        return pyeelight.Bulb.connect(self.get_ip())


def get_bulbs():
    ad_socket = AdvertisementSocket()
    ad_socket.send_packet(OutboundRequestPacket("ssdp:discover", "wifi_bulb"))

    return [BulbInfo(i.name, i.location) for i in ad_socket.init_waiter_thread()]


import json
import socket
from abc import abstractmethod


class Contextable:

    @abstractmethod
    def get_context(self):
        pass


class Logger:

    def __init__(self, ctx):
        self.ctx = ctx

    def batch_received(self, messages):
        for message in messages.decode().strip().split("\r\n"):
            if "id" in json.loads(message):
                self.received(message)
            else:
                self.generic(message)

    def generic(self, message):
        received_obj = json.loads(message)
        final_msg = []
        for prop in received_obj:
            final_msg.append(f"{prop}: {received_obj[prop]}")
        print(f"[{self.ctx.get_context()}][GENERIC] {', '.join(final_msg)}")

    def received(self, message):
        try:
            received_obj = json.loads(message)
            self.info(received_obj["id"], ", ".join(received_obj["result"]))
        except:
            print("ERROR:", message)

    @staticmethod
    def dispatch(func):
        def wrapper(*args):
            bulb_obj = args[0]
            params = args[1:]
            bulb_obj.logger.info(bulb_obj.formatter.id + 1,
                                 f"Sent {func.__name__}({' '.join(map(lambda x: str(x), params))}) command to bulb")
            func(*args)
            bulb_obj.logger.batch_received(bulb_obj.socket_conn.recv(1024))

        return wrapper

    def sent(self, command):
        pass

    def info(self, protocol_id, message):
        print(f"[{self.ctx.get_context()}][{protocol_id}] {message}")


class Bulb(Contextable):

    def __init__(self, socket_conn):
        self.socket_conn = socket_conn
        self.formatter = CommandFormatter()
        self.logger = Logger(self)

    @staticmethod
    def connect(ip):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((ip, 55443))
        print("Connected to bulb!")
        return Bulb(s)

    @Logger.dispatch
    def set_power(self, state):
        self.socket_conn.sendall(self.formatter.format_command("set_power", "on" if state else "off").encode())

    @Logger.dispatch
    def set_brightness(self, brightness):
        self.socket_conn.sendall(self.formatter.format_command("set_bright", brightness).encode())

    @Logger.dispatch
    def get_properties(self, *properties):
        self.socket_conn.sendall(self.formatter.format_command("get_prop", *properties).encode())

    @Logger.dispatch
    def toggle(self):
        self.socket_conn.sendall(self.formatter.format_command("toggle").encode())

    @Logger.dispatch
    def set_rgb(self, r, g, b):
        self.socket_conn.sendall(self.formatter.format_command("set_rgb", ((r * 65536) + (g * 256) + b)).encode())

    @Logger.dispatch
    def reset(self):
        self.socket_conn.sendall(self.formatter.format_command("set_default").encode())

    @Logger.dispatch
    def set_name(self, name):
        self.socket_conn.sendall(self.formatter.format_command("set_name", name).encode())

    def get_context(self):
        return f"BULB::{self.socket_conn.getpeername()[0]}"


class CommandFormatter:

    def __init__(self):
        self.id = 0

    def format_command(self, method, *params):
        self.id += 1
        return json.dumps({"id": self.id, "method": method, "params": params}) + "\r\n"




from setuptools import setup, find_packages

setup(
    name='pyeelight',
    version='0.4.0',
    license='MIT',
    author="Mohamed El Yousfi",
    author_email='pip@mootje.be',
    packages=find_packages(),
    url='https://github.com/Mohagames205/pyeelight-api',
    keywords=["yeelight", "api", "smart home"],
)